#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron.api import extensions
from neutron.api.v2 import attributes as attr
from neutron.api.v2 import resource_helper
from neutron.plugins.common import constants
from neutron.services import service_base
from neutron_lib.api import converters as conv
from neutron_lib.api import validators as valid
from neutron_lib import exceptions as nexc
from oslo_config import cfg
from oslo_log import log as logging
import six

from gbpservice.neutron import extensions as gbp_extensions
import gbpservice.neutron.extensions.group_policy  # noqa
from gbpservice.neutron.services.servicechain.common import constants as scc


# The code below is a monkey patch of key Neutron's modules. This is needed for
# the GBP service to be loaded correctly. GBP extensions' path is added
# to Neutron's so that it's found at extension scanning time.
extensions.append_api_extensions_path(gbp_extensions.__path__)
LOG = logging.getLogger(__name__)


# Service Chain Exceptions
class ServiceProfileNotFound(nexc.NotFound):
    message = _("ServiceProfile %(profile_id)s could not be found")


class ServiceProfileInUse(nexc.NotFound):
    message = _("Unable to complete operation, ServiceProfile "
                "%(profile_id)s is in use")


class ServiceChainNodeNotFound(nexc.NotFound):
    message = _("ServiceChainNode %(sc_node_id)s could not be found")


class ServiceChainSpecNotFound(nexc.NotFound):
    message = _("ServiceChainSpec %(sc_spec_id)s could not be found")


class ServiceChainInstanceNotFound(nexc.NotFound):
    message = _("ServiceChainInstance %(sc_instance_id)s could not be found")


class ServiceChainNodeInUse(nexc.InUse):
    message = _("Unable to complete operation, ServiceChainNode "
                "%(node_id)s is in use")


class ServiceChainSpecInUse(nexc.InUse):
    message = _("Unable to complete operation, ServiceChainSpec "
                "%(spec_id)s is in use")


class ServiceTypeNotFound(nexc.NotFound):
    message = _("ServiceType %(service_type_id) could not be found")


class ServiceTypeNotSupported(nexc.NotFound):
    message = _("ServiceType %(service_type_id) not supported")


class PortNotFound(nexc.NotFound):
    message = _("Port %(port_id)s could not be found")


def _validate_str_list(data, valid_values=None):
    if not isinstance(data, list):
        msg = _("'%s' is not a list") % data
        LOG.debug(msg)
        return msg

    for item in data:
        msg = valid.validate_string(item)
        if msg:
            LOG.debug(msg)
            return msg

    if len(set(data)) != len(data):
        msg = _("Duplicate items in the list: '%s'") % ', '.join(data)
        LOG.debug(msg)
        return msg


valid.validators['type:string_list'] = _validate_str_list

SERVICECHAIN_NODES = 'servicechain_nodes'
SERVICECHAIN_SPECS = 'servicechain_specs'
SERVICECHAIN_INSTANCES = 'servicechain_instances'
SERVICE_PROFILES = 'service_profiles'

RESOURCE_ATTRIBUTE_MAP = {
    SERVICECHAIN_NODES: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None}, 'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None}, 'default': '',
                 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'service_type': {'allow_post': True, 'allow_put': False,
                         'validate': {'type:string_or_none': None},
                         'is_visible': True, 'default': None},
        'service_profile_id': {'allow_post': True, 'allow_put': True,
                               'validate': {'type:uuid_or_none': None},
                               'is_visible': True, 'default': None},
        'config': {'allow_post': True, 'allow_put': True,
                   'validate': {'type:string': None},
                   'required': True, 'is_visible': True},
        attr.SHARED: {'allow_post': True, 'allow_put': True,
                      'default': False, 'convert_to': conv.convert_to_boolean,
                      'is_visible': True, 'required_by_policy': True,
                      'enforce_policy': True},
    },
    SERVICECHAIN_SPECS: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None}, 'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'nodes': {'allow_post': True, 'allow_put': True,
                  'validate': {'type:uuid_list': None},
                  'convert_to': conv.convert_none_to_empty_list,
                  'default': None, 'is_visible': True,
                  'required': True},
        'config_param_names': {'allow_post': False, 'allow_put': False,
                               'validate': {'type:string_list': None},
                               'default': [], 'is_visible': True},
        attr.SHARED: {'allow_post': True, 'allow_put': True,
                      'default': False, 'convert_to': conv.convert_to_boolean,
                      'is_visible': True, 'required_by_policy': True,
                      'enforce_policy': True},
    },
    SERVICECHAIN_INSTANCES: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None}, 'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        'servicechain_specs': {'allow_post': True, 'allow_put': True,
                              'validate': {'type:uuid_list': None},
                              'convert_to': conv.convert_none_to_empty_list,
                              'default': None, 'is_visible': True,
                              'required': True},
        'provider_ptg_id': {'allow_post': True, 'allow_put': False,
                            'validate': {'type:uuid_or_none': None},
                            'is_visible': True, 'default': None,
                            'required': True},
        'consumer_ptg_id': {'allow_post': True, 'allow_put': False,
                            'validate': {'type:string_or_none': None},
                            'is_visible': True, 'default': None,
                            'required': True},
        'management_ptg_id': {'allow_post': True, 'allow_put': False,
                              'validate': {'type:uuid_or_none': None},
                              'is_visible': True, 'default': None,
                              'required': True},
        'classifier_id': {'allow_post': True, 'allow_put': True,
                          'validate': {'type:uuid_or_none': None},
                          'is_visible': True, 'default': None,
                          'required': True},
        'config_param_values': {'allow_post': True, 'allow_put': False,
                                'validate': {'type:string': None},
                                'default': "", 'is_visible': True},
    },
    SERVICE_PROFILES: {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:uuid': None}, 'is_visible': True,
               'primary_key': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:gbp_resource_name': None},
                 'default': '', 'is_visible': True},
        'description': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True, 'default': ''},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'required_by_policy': True, 'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'status_details': {'allow_post': False, 'allow_put': False,
                           'is_visible': True},
        attr.SHARED: {'allow_post': True, 'allow_put': True,
                      'default': False, 'convert_to': conv.convert_to_boolean,
                      'is_visible': True, 'required_by_policy': True,
                      'enforce_policy': True},
        'vendor': {'allow_post': True, 'allow_put': True,
                   'validate': {'type:string': None},
                   'is_visible': True, 'default': ''},
        'insertion_mode': {'allow_post': True, 'allow_put': True,
                           'validate': {'type:values':
                                        scc.VALID_INSERTION_MODES},
                           'is_visible': True, 'default': None},
        'service_type': {'allow_post': True, 'allow_put': True,
                         'validate': {'type:string': None},
                         'is_visible': True, 'required': True},
        'service_flavor': {'allow_post': True, 'allow_put': True,
                           'validate': {'type:string_or_none': None},
                           'is_visible': True, 'default': None},
    },
}


service_chain_quota_opts = [
    cfg.IntOpt('quota_servicechain_node',
               default=-1,
               help=_('Number of Service Chain Nodes allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_servicechain_spec',
               default=-1,
               help=_('Number of Service Chain Specs allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_servicechain_instance',
               default=-1,
               help=_('Number of Service Chain Instances allowed per tenant. '
                      'A negative value means unlimited.')),
    cfg.IntOpt('quota_service_profile',
               default=-1,
               help=_('Number of Service Profiles allowed per tenant. '
                      'A negative value means unlimited.')),
]
cfg.CONF.register_opts(service_chain_quota_opts, 'QUOTAS')


class Servicechain(extensions.ExtensionDescriptor):

    @classmethod
    def get_name(cls):
        return "Service Chain Abstraction"

    @classmethod
    def get_alias(cls):
        return "servicechain"

    @classmethod
    def get_description(cls):
        return "Extension for Service Chain Abstraction"

    @classmethod
    def get_namespace(cls):
        return "http://wiki.openstack.org/neutron/sc/v2.0/"

    @classmethod
    def get_updated(cls):
        return "2014-08-03T12:00:00-00:00"

    @classmethod
    def get_resources(cls):
        plural_mappings = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        gbp_extensions.register_plurals(plural_mappings)
        return resource_helper.build_resource_info(plural_mappings,
                                                   RESOURCE_ATTRIBUTE_MAP,
                                                   constants.SERVICECHAIN,
                                                   register_quota=True)

    @classmethod
    def get_plugin_interface(cls):
        return ServiceChainPluginBase

    def update_attributes_map(self, attributes):
        super(Servicechain, self).update_attributes_map(
            attributes, extension_attrs_map=RESOURCE_ATTRIBUTE_MAP)

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}


@six.add_metaclass(abc.ABCMeta)
class ServiceChainPluginBase(service_base.ServicePluginBase):

    def get_plugin_name(self):
        return constants.SERVICECHAIN

    def get_plugin_type(self):
        return constants.SERVICECHAIN

    def get_plugin_description(self):
        return 'Service Chain plugin'

    def update_chains_pt_added(self, context, policy_target, instance_id):
        """ Auto scaling function.

        Override this method to react to policy target creation.
        """
        pass

    def update_chains_pt_removed(self, context, policy_target, instance_id):
        """ Auto scaling function.

        Override this method to react to policy target deletion.
        """
        pass

    def update_chains_consumer_added(self, context, policy_target_group,
                                     instance_id):
        """ Auto scaling function.

        Override this method to react to policy target group addition as
        a consumer of a chain.
        """
        pass

    def update_chains_consumer_removed(self, context, policy_target_group,
                                       instance_id):
        """ Auto scaling function.

        Override this method to react to policy target group removed as a
        consumer of a chain
        """
        pass

    def policy_target_group_updated(self, context, old_policy_target_group,
                                    current_policy_target_group,
                                    instance_id):
        """ Utility function.

        Override this method to react to policy target group update
        """
        pass

    @abc.abstractmethod
    def get_servicechain_nodes(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_servicechain_node(self, context, servicechain_node_id,
                              fields=None):
        pass

    @abc.abstractmethod
    def create_servicechain_node(self, context, servicechain_node):
        pass

    @abc.abstractmethod
    def update_servicechain_node(self, context, servicechain_node_id,
                                 servicechain_node):
        pass

    @abc.abstractmethod
    def delete_servicechain_node(self, context, servicechain_node_id):
        pass

    @abc.abstractmethod
    def get_servicechain_specs(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_servicechain_spec(self, context, servicechain_spec_id,
                              fields=None):
        pass

    @abc.abstractmethod
    def create_servicechain_spec(self, context, servicechain_spec):
        pass

    @abc.abstractmethod
    def update_servicechain_spec(self, context, servicechain_spec_id,
                                 servicechain_spec):
        pass

    @abc.abstractmethod
    def delete_servicechain_spec(self, context, servicechain_spec_id):
        pass

    @abc.abstractmethod
    def get_servicechain_instances(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_servicechain_instance(self, context, servicechain_instance_id,
                                  fields=None):
        pass

    @abc.abstractmethod
    def create_servicechain_instance(self, context, servicechain_instance):
        pass

    @abc.abstractmethod
    def update_servicechain_instance(self, context, servicechain_instance_id,
                                     servicechain_instance):
        pass

    @abc.abstractmethod
    def delete_servicechain_instance(self, context, servicechain_instance_id):
        pass

    @abc.abstractmethod
    def create_service_profile(self, context, service_profile):
        pass

    @abc.abstractmethod
    def update_service_profile(self, context, service_profile_id,
                               service_profile):
        pass

    @abc.abstractmethod
    def delete_service_profile(self, context, service_profile_id):
        pass

    @abc.abstractmethod
    def get_service_profile(self, context, service_profile_id, fields=None):
        pass

    @abc.abstractmethod
    def get_service_profiles(self, context, filters=None, fields=None):
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class FakeObjects(object):
    """ Implements fake objects for assertion.

    """

    sc = 'sc'
    empty_dict = {}
    context = 'APIcontext'
    neutron_context = {
        'agent_info': {
            'service_type': 'firewall',
            'notification_data': {},
            'service_vendor': 'vyos',
            'resource': 'firewall',
            'context': 'APIcontext'},
        'neutron context for *aaS': {}}
    firewall = {'id': 'firewall'}
    host = 'host'
    conf = 'conf'
    vmid = 'b238e3f12fb64ebcbda2b3330700bf00'
    rpcmgr = 'rpcmgr'
    drivers = 'drivers'
    data_for_interface = dict(provider_mac="fa:16:3e:d9:4c:33",
                              stitching_mac="fa:16:3e:da:ca:4d")
    data_for_add_src_route = [{'source_cidr': "11.0.1.0/24",
                              'gateway_ip': "192.168.0.1"},
                              {'source_cidr': "192.168.0.0/28",
                              'gateway_ip': "192.168.0.1"}]
    data_for_del_src_route = [{'source_cidr': '11.0.1.0/24'},
                              {'source_cidr': '192.168.0.0/28'}]
    timeout = 180

    def get_url_for_api(self, api):
        url = 'http://11.0.0.37:8888/'
        api_url_map = {
            'log_forward': 'configure-rsyslog-as-client',
            'add_static_ip': 'add_static_ip',
            'add_inte': 'add_rule',
            'del_inte': 'delete_rule',
            'add_src_route': 'add-source-route',
            'del_src_route': 'delete-source-route',
            'config_fw': 'configure-firewall-rule',
            'update_fw': 'update-firewall-rule',
            'delete_fw': 'delete-firewall-rule'}

        return url + api_url_map[api]

    def log_forward_data(self):
        return dict(server_ip={}, server_port={}, log_level={})

    def static_ip_data(self):
        return dict(
                    provider_ip="11.0.1.1",
                    provider_cidr="11.0.1.0/24",
                    provider_mac="fa:16:3e:d9:4c:33",
                    stitching_ip="192.168.0.3",
                    stitching_cidr="192.168.0.0/28",
                    stitching_mac="fa:16:3e:da:ca:4d")

    def firewall_api_context(self):
        fw_context = {
            'agent_info': {
                'resource': 'firewall',
                'service_vendor': 'vyos',
                'context': {'requester': 'device_orch',
                            'logging_context': {}},
                'resource_type': 'firewall'},
            'notification_data': {}, 'service_info': {},
            "resource_data": {
                    "forward_route": True,
                    "tenant_id": "ac33b4c2d80f485a86ea515c09c74949",
                    "nfs": [{
                        "role": "master",
                        "svc_mgmt_fixed_ip": "11.0.0.37",
                        "networks": [
                            {"cidr": "11.0.1.0/24",
                             "gw_ip": "",
                             "type": "provider",
                             "ports": [{
                                "mac": "fa:16:3e:d9:4c:33",
                                "fixed_ip": "11.0.1.1",
                                "floating_ip": ""}]},
                            {"cidr": "192.168.0.0/28",
                             "gw_ip": "192.168.0.1 ",
                             "type": "stitching",
                             "ports": [{
                                 "mac": "fa:16:3e:da:ca:4d",
                                 "fixed_ip": "192.168.0.3",
                                 "floating_ip": ""}]}
                                        ]}]}}
        return fw_context

    def fake_request_data_generic_bulk(self):
        """ A sample bulk request data for generic APIs

        Returns: data which is the input for generic configuration
        RPC receivers of configurator.

        """

        request_data = {
            "info": {
                "service_type": "firewall",
                "service_vendor": "vyos",
                "context": {
                        "requester": "device_orch",
                        "logging_context": {},
                        "nf_id": "nf_id",
                        "nfi_id": "nfi_id"
                }
            },
            "config": [{
                "resource": "interfaces",
                "resource_data": {
                    "forward_route": True,
                    "tenant_id": "ac33b4c2d80f485a86ea515c09c74949",
                    "nfds": [{
                        "role": "master",
                        "svc_mgmt_fixed_ip": "11.0.0.37",
                        "networks": [
                            {"cidr": "11.0.1.0/24",
                             "gw_ip": "",
                             "type": "provider",
                             "ports": [{
                                "mac": "fa:16:3e:d9:4c:33",
                                "fixed_ip": "11.0.1.1",
                                "floating_ip": ""}]},
                            {"cidr": "192.168.0.0/28",
                             "gw_ip": "192.168.0.1 ",
                             "type": "stitching",
                             "ports": [{
                                 "mac": "fa:16:3e:da:ca:4d",
                                 "fixed_ip": "192.168.0.3",
                                 "floating_ip": ""}]}
                                        ]}]}},
                       {
                "resource": "routes",
                "resource_data": {
                    "forward_route": True,
                    "tenant_id": "ac33b4c2d80f485a86ea515c09c74949",
                    "nfds": [{
                        "role": "master",
                        "svc_mgmt_fixed_ip": "11.0.0.37",
                        "networks": [
                            {"cidr": "11.0.1.0/24",
                             "gw_ip": "",
                             "type": "provider",
                             "ports": [{
                                "mac": "fa:16:3e:d9:4c:33",
                                "fixed_ip": "11.0.1.1",
                                "floating_ip": ""}]},
                            {"cidr": "192.168.0.0/28",
                             "gw_ip": "192.168.0.1 ",
                             "type": "stitching",
                             "ports": [{
                                 "mac": "fa:16:3e:da:ca:4d",
                                 "fixed_ip": "192.168.0.3",
                                 "floating_ip": ""}]}
                                        ]}]}}]
        }
        return request_data

    def fake_request_data_generic_single(self, routes=False):
        """ A sample single request data for generic APIs

        Returns: data which is the input for generic configuration
        RPC receivers of configurator.

        """

        request_data = self.fake_request_data_generic_bulk()
        request_data['config'].pop(0) if routes else (
                                            request_data['config'].pop(1))

        return request_data

    def fake_request_data_fw(self):
        """ A sample request data for FwaaS APIs

        Returns: data which is the input for firewall configuration
        RPC receivers of configurator.

        """

        request_data = {
            "info": {
                "service_type": "firewall",
                "service_vendor": "vyos",
                "context": {
                        "requester": "device_orch",
                        "logging_context": {}
                }
            },
            "config": [{
                "resource": "firewall",
                "resource_data": {
                    "neutron_context": {
                                    "notification_data": {},
                                    "resource": "firewall"
                                },
                    "firewall": self._fake_firewall_obj(),
                    "host": self.host}}]
                        }
        return request_data

    def fake_sa_req_list_fw(self):
        """ A sample data for firewall agent handlers

        Returns: data which is the input for event handler
        functions of firewall agent.

        """

        request_data = [{
                        "agent_info": {
                            "service_vendor": "vyos",
                            "service_feature": "",
                            "resource": "firewall",
                            "context": {
                                "requester": "device_orch",
                                "logging_context": {}
                            },
                            "resource_type": "firewall"
                        },
                        "method": "create_firewall",
                        "resource_data": {
                            "firewall": {
                                "status": "PENDING_CREATE",
                                "router_ids": [
                                    "650bfd2f-7766-4a0d-839f-218f33e16998"
                                ],
                                "description": '{\
                                    "vm_management_ip": "172.24.4.5",\
                                    "provider_cidr": "11.0.1.0/24",\
                                    "service_vendor": "vyos"}',
                                "admin_state_up": True,
                                "firewall_policy_id":
                                    "c69933c1-b472-44f9-8226-30dc4ffd454c",
                                "tenant_id":
                                    "45977fa2dbd7482098dd68d0d8970117",
                                "id": "3b0ef8f4-82c7-44d4-a4fb-6177f9a21977",
                                "firewall_rule_list": True,
                                "name": ""
                            },
                            "neutron_context": {
                                "notification_data": {},
                                "resource": "firewall"
                            },
                            "host": "host"
                        },
                        "is_generic_config": False
                        }]

        return request_data

    def fake_sa_req_list(self):
        """ A sample data for generic config agent handlers

        Returns: data which is the input for generic config event
        handler functions of agents.

        """

        sa_req_list = [{
            'agent_info': {
                'service_vendor': 'vyos',
                'service_feature': '',
                'resource': 'interfaces',
                'resource_type': 'firewall',
                'context': {
                    'logging_context': {},
                    'requester': 'device_orch'
                }
            },
            'method': 'configure_interfaces',
            'resource_data': {
                'forward_route': True,
                'tenant_id': 'ac33b4c2d80f485a86ea515c09c74949',
                'nfds': [{
                    'role': 'master',
                    'networks': [{
                        'cidr': '11.0.1.0/24',
                        'gw_ip': '',
                        'type': 'provider',
                        'ports': [{
                            'fixed_ip': '11.0.1.1',
                            'mac': 'fa:16:3e:d9:4c:33',
                            'floating_ip': ''
                        }]
                    }, {
                        'cidr': '192.168.0.0/28',
                        'gw_ip': '192.168.0.1 ',
                        'type': 'stitching',
                        'ports': [{
                            'mac': 'fa:16:3e:da:ca:4d',
                            'floating_ip': '',
                            'fixed_ip': '192.168.0.3'
                        }]
                    }],
                    'svc_mgmt_fixed_ip': '11.0.0.37'
                }]
            },
            'is_generic_config': True
        }, {
            'agent_info': {
                'service_vendor': 'vyos',
                'service_feature': '',
                'resource': 'routes',
                'resource_type': 'firewall',
                'context': {
                    'logging_context': {},
                    'requester': 'device_orch'
                }
            },
            'method': 'configure_routes',
            'resource_data': {
                'forward_route': True,
                'tenant_id': 'ac33b4c2d80f485a86ea515c09c74949',
                'nfds': [{
                    'role': 'master',
                    'networks': [{
                        'cidr': '11.0.1.0/24',
                        'gw_ip': '',
                        'type': 'provider',
                        'ports': [{
                            'fixed_ip': '11.0.1.1',
                            'mac': 'fa:16:3e:d9:4c:33',
                            'floating_ip': ''
                        }]
                    }, {
                        'cidr': '192.168.0.0/28',
                        'gw_ip': '192.168.0.1 ',
                        'type': 'stitching',
                        'ports': [{
                            'mac': 'fa:16:3e:da:ca:4d',
                            'floating_ip': '',
                            'fixed_ip': '192.168.0.3'
                        }]
                    }],
                    'svc_mgmt_fixed_ip': '11.0.0.37'
                }]
            },
            'is_generic_config': True}]

        return sa_req_list

    def _fake_resource_data(self):
        """ A sample keyword arguments for configurator

        Returns: kwargs

        """

        resource_data = {
                'forward_route': True,
                'tenant_id': 'ac33b4c2d80f485a86ea515c09c74949',
                'fail_count': 0,
                'nfds': [{
                    'role': 'master',
                    'networks': [{
                        'cidr': '11.0.1.0/24',
                        'gw_ip': '',
                        'type': 'provider',
                        'ports': [{
                            'fixed_ip': '11.0.1.1',
                            'mac': 'fa:16:3e:d9:4c:33',
                            'floating_ip': ''
                        }]
                    }, {
                        'cidr': '192.168.0.0/28',
                        'gw_ip': '192.168.0.1',
                        'type': 'stitching',
                        'ports': [{
                            'mac': 'fa:16:3e:da:ca:4d',
                            'floating_ip': '',
                            'fixed_ip': '192.168.0.3'
                        }]
                    }],
                    'svc_mgmt_fixed_ip': '11.0.0.37',
                    'periodicity': 'initial',
                    'vmid': 'b238e3f12fb64ebcbda2b3330700bf00'
                }]}

        return resource_data

    def _fake_firewall_obj(self):
        """ A sample firewall resource object

        Returns: firewall object

        """

        firewall = {
                     "admin_state_up": True,
                     "description": "",
                     "firewall_policy_id": (
                                "c69933c1-b472-44f9-8226-30dc4ffd454c"),
                     "id": "3b0ef8f4-82c7-44d4-a4fb-6177f9a21977",
                     "name": "",
                     "status": "PENDING_CREATE",
                     "router_ids": [
                         "650bfd2f-7766-4a0d-839f-218f33e16998"
                     ],
                     "tenant_id": "45977fa2dbd7482098dd68d0d8970117",
                     "description": '{\
                                    "vm_management_ip": "172.24.4.5",\
                                    "provider_cidr": "11.0.1.0/24",\
                                    "service_vendor": "vyos"}',
                     "firewall_rule_list": True
                    }
        return firewall


class FakeEventFirewall(object):
    """ Implements a fake event class for firewall for
        process framework to use

    """

    def __init__(self):
        fo = FakeObjects()
        kwargs = fo._fake_resource_data()
        self.data = {
                    'context': {'neutron context for *aaS': {},
                                'agent_info': {
                                        'resource': 'healthmonitor',
                                        'notification_data': {},
                                        'service_type': 'firewall',
                                        'service_vendor': 'vyos',
                                        'context': 'APIcontext'}},
                    'firewall': fo._fake_firewall_obj(),
                    'host': fo.host,
                    'resource_data': kwargs}
        self.id = 'dummy'


class FakeEventGenericConfig(object):
    """ Implements a fake event class for generic config for
        process framework to use

    """

    def __init__(self):
        fo = FakeObjects()
        kwargs = fo._fake_resource_data()
        self.data = {
                    'context': {
                            'resource': 'healthmonitor',
                            'notification_data': {},
                            'resource_type': 'firewall',
                            'service_vendor': 'vyos',
                            'context': 'APIcontext'},
                    'firewall': fo._fake_firewall_obj(),
                    'host': fo.host,
                    'resource_data': kwargs}
        self.id = 'dummy'


class FakeEventGetNotifications(object):
    """ Implements a fake event class for notifications functionality
        for the process framework to use

    """

    def __init__(self):
        self.data = {'dummy_data': 'dummy_value'}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import six
import yaml

from gbpservice.nfp.common import constants as nfp_constants
from gbpservice.nfp.core import log as nfp_logging

LOG = nfp_logging.getLogger(__name__)

NEUTRON_ML2_CONF = "/etc/neutron/plugins/ml2/ml2_conf.ini"


def _parse_service_flavor_string(service_flavor_str):
    service_details = {}
    if ',' not in service_flavor_str:
        service_details['device_type'] = 'nova'
        service_details['service_vendor'] = service_flavor_str
    else:
        service_flavor_dict = dict(item.split('=') for item
                                   in service_flavor_str.split(','))
        service_details = {key.strip(): value.strip() for key, value
                           in six.iteritems(service_flavor_dict)}
        return service_details


def _get_dict_desc_from_string(vpn_svc):
    svc_desc = vpn_svc.split(";")
    desc = {}
    for ele in svc_desc:
        s_ele = ele.split("=")
        desc.update({s_ele[0]: s_ele[1]})
    return desc


def get_vpn_description_from_nf(network_function):
    str_description = network_function['description'].split('\n')[1]
    description = _get_dict_desc_from_string(
            str_description)
    return description, str_description


def is_vpn_in_service_chain(sc_specs):
    for spec in sc_specs:
        nodes = spec['sc_nodes']
        for node in nodes:
            service_type = node['sc_service_profile']['service_type']
            if service_type.lower() == nfp_constants.VPN:
                return True
    return False


def get_config_file(dir_name, service_vendor):
    file_name = ''
    for _file in os.listdir(dir_name):
        if _file.startswith(service_vendor):
            file_name = _file
            break
    return file_name


def get_service_vm_context(service_vendor, tenant_name=None):
        """ Load day0 config file
            :param service_vendor: service vendor name
            :param tenant_name

            - Day0 file name must start with service vendor name followed by
              string '_day0'
              e.g Vyos day0 file name can be vyos_day0.json or vyos_day0
            - File format can be of any type like text file, json file etc

            - service vendor specific default day0 config file
                /etc/nfp/<service_vendor>/<day0_file>
                e.g /etc/nfp/vyos/vyos_day0.json
            - tenant specific vendor day0 config file
                /etc/nfp/<service_vendor>/<tenant_name>/<day0_file>
                e.g /etc/nfp/vyos/services/vyos_day0.json

            Returns - day0 config file
        """
        try:
            file_name = ''
            default_config_dir = nfp_constants.CONFIG_DIR
            vendor_day0_dir = default_config_dir + service_vendor + '/'
            if tenant_name:
                tenant_day0_dir = vendor_day0_dir + tenant_name + '/'
                if os.path.isdir(tenant_day0_dir):
                    file_name = get_config_file(tenant_day0_dir,
                                                service_vendor)
            if file_name:
                day0_config_file = tenant_day0_dir + file_name
            else:
                if os.path.isdir(vendor_day0_dir):
                    file_name = get_config_file(vendor_day0_dir,
                                                service_vendor)
                    day0_config_file = vendor_day0_dir + file_name
                else:
                    day0_config_file = '/fake_file_path'

            with open(day0_config_file) as _file:
                try:
                    svm_context = yaml.load(_file)
                except Exception as e:
                    msg = ("Failed yaml load file %s. Reason: %s"
                           % (day0_config_file, e))
                    raise Exception(msg)

            msg = ("Loaded day0 config file %s for service_vendor %s,"
                   "tenant_name %s" % (day0_config_file, service_vendor,
                                       tenant_name))
            LOG.info(msg)
            return svm_context
        except Exception as ex:
            msg = ("Failed to read day0 config file, ERROR: %s" % ex)
            LOG.error(msg)
            return None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import model_base
import sqlalchemy as sa


class ApicAllowedVMNameDB(model_base.BASEV2):
    __tablename__ = 'gp_apic_mapping_allowed_vm_names'
    l3_policy_id = sa.Column(sa.String(36),
                             sa.ForeignKey('gp_l3_policies.id',
                                           ondelete="CASCADE"),
                             primary_key=True)
    allowed_vm_name = sa.Column(sa.String(255), primary_key=True)


class ApicAllowedVMNameDBMixin(object):

    def get_l3_policy_allowed_vm_names(self, session, l3_policy_id):
        rows = (session.query(ApicAllowedVMNameDB).filter_by(
                l3_policy_id=l3_policy_id).all())
        return rows

    def get_l3_policy_allowed_vm_name(self, session, l3_policy_id,
                                      allowed_vm_name):
        row = (session.query(ApicAllowedVMNameDB).filter_by(
            l3_policy_id=l3_policy_id,
            allowed_vm_name=allowed_vm_name).one())
        return row

    def add_l3_policy_allowed_vm_name(self, session, l3_policy_id,
                                      allowed_vm_name):
        row = ApicAllowedVMNameDB(l3_policy_id=l3_policy_id,
                                  allowed_vm_name=allowed_vm_name)
        session.add(row)

    def delete_l3_policy_allowed_vm_name(self, session, l3_policy_id,
                                         allowed_vm_name):
        row = self.get_l3_policy_allowed_vm_name(
            session, l3_policy_id, allowed_vm_name)
        session.delete(row)
